from abc import abstractmethod, ABCMeta

class AbstractCircuitBuilder(metaclass=ABCMeta):
    def __init__(self):
        pass
    @abstractmethod
    def __call__(self, *args):
        pass
class AbstractSingleGateCircuitBuilder(AbstractCircuitBuilder):
    def __init__(self):
        self._registry = dict()
    def clear_gate_registry(self):
        self._registry = dict()



import pytest
import ray
import numpy as np

from pyqcs.experiment.gradient_decent import approx_gradient_decent
from pyqcs.experiment.workflow import WorkflowSpawner, FunctionInstruction

@pytest.mark.slow
def test_gradient_decent_convergence(ray_setup):
    f = lambda x: np.sum(x**2 + x**6)
    x0 = np.array(list(range(3)))
    instructions = [FunctionInstruction("quadratic function", f)]
    wf_spawner = WorkflowSpawner("quadratic function evaluation", instructions)

    success, result = approx_gradient_decent(wf_spawner, x0, 1e-5, 2e-3)

    assert success is True
    assert np.allclose(result, np.zeros_like(x0), atol=1e-5)



from .. import list_to_circuit
from ..gates.builtins import H, S, CZ, Z, X
from ..graph.state import GraphState
from ..graph.c_l import C_L

def graph_state_to_circuit(state):
    """
    This function basically converts a graph state to a circuit
    that (applied to a zero-state) will yield the same state.

    In particular for some graph state g::

        graph_state_to_circuit(g) * GraphState.new_zero_state(g._nbits)

    will copy g.
    """

    if(not isinstance(state, GraphState)):
        raise TypeError(f"GraphState is required, but got {type(state)}")

    prepare = list_to_circuit([H(i) for i in range(state._nbits)])

    vops, edges = state._g_state.to_lists()

    entanglements = []

    handled_edges = set()
    for i, ngbhd in enumerate(edges):
        for j in ngbhd:
            edge = tuple(sorted((i, j)))
            if(edge in handled_edges):
                continue
            entanglements.append(CZ(*edge))
            handled_edges |= {edge}

    entanglements = list_to_circuit(entanglements)

    vop_circuit = list_to_circuit([vop_to_circuit(i, vop) for i,vop in enumerate(vops)])

    return prepare | entanglements | vop_circuit

def vop_to_circuit(act, vop):
    s2m = {"H": H, "S": S, "X": X, "Z": Z}
    return list_to_circuit([s2m[c](act) for c in reversed(C_L[vop])])



from .gates.builtins import M
from .utils import list_to_circuit
from .state.state import BasicState as State
from collections import Counter, deque

import numpy as np

def build_measurement_circuit(bit_mask):
    if(isinstance(bit_mask, int)):
        c_list = [M(i) for i in range(bit_mask.bit_length()) if bit_mask & (1 << i)]
    elif(isinstance(bit_mask, (list, tuple))):
        c_list = [M(i) for i in bit_mask]
    else:
        raise TypeError("bit_mask must be either int, list or tuple, but {} is given".format(type(bit_mask)))
    return list_to_circuit(c_list)


def measure(state, bit_mask):
    """
    Measures the qbits ``bit_mask`` (in the order given by
    ``bit_mask``, if bit_mask is a list, or least significant
    bit first).

    The original state is unchanged.

    Returns ``new_state, bit_string: int``.
    """
    circuit = build_measurement_circuit(bit_mask)

    state = state.deepcopy()
    if(isinstance(state, State)):
        state._cl_state[:] = -1
    else:
        state._measured = dict()
    new_state = circuit * state
    if(isinstance(state, State)):
        return new_state, sum([1 << i for i,v in enumerate(new_state._cl_state) if v == 1])
    else:
        return new_state, sum([1 << i for i,v in new_state._measured.items() if v == 1])


def _do_sample(state, circuit, nsamples):
    for _ in range(nsamples):
        new_state = circuit * state
        if(isinstance(new_state, State)):
            yield new_state, sum([1 << i for i,v in enumerate(new_state._cl_state) if v == 1])
        else:
            yield new_state, sum([1 << i for i,v in new_state._measured.items() if v == 1])

def sample(state, bit_mask, nsamples, keep_states=False):
    """
    Measures the qbits given in ``bit_mask`` ``nsamples`` times
    and returns a counter ``{result: count}``.

    If ``keep_states is True`` the resulting states are included in ``result``.
    This does not work for graphical states because there is currently no
    meaningful way to hash graphical states.

    The original state is unchanged.
    """
    circuit = build_measurement_circuit(bit_mask)

    state = state.deepcopy(force_new_state=True)
    if(isinstance(state, State)):
        state._cl_state[:] = -1
    else:
        state._measured = dict()

    if(keep_states):
        return Counter(_do_sample(state, circuit, nsamples))

    return Counter((i[1] for i in _do_sample(state, circuit, nsamples)))

def tree_amplitudes(state, bit_mask=None, eps=1e-5):
    """
    Compute the probability amplitudes for all (eps-)possible
    outcomes. ``bit_mask`` is either ``None`` or a permutation
    of ``list(range(state._nbits))``.

    Only available for dense vector states.

    The original state is unchanged.

    Amplitudes (and collapsed states) are computed in the order given
    by ``bit_mask``. If ``bit_mask is None``, ``list(range(state._nbits))``
    is used.
    """

    if(not isinstance(state, State)):
        raise TypeError("tree_amplitudes currently works for dense vector states only")

    if(bit_mask is None):
        bit_mask = list(range(state._nbits))

    if(list(sorted(bit_mask)) != list(range(state._nbits))):
        raise ValueError("bit_mask must be either None or a permutation of list(range(state._nbits)))")

    next_queue = [(1, 0, state.deepcopy()._qm_state)]
    qbit_mapping = np.arange(0, 2**state._nbits, 1, dtype=np.int)

    for qbit in bit_mask:
        this_queue = next_queue
        next_queue = deque()
        while(this_queue):
            prob, prev_outcome, handle_now = this_queue.pop()

            bit_mask_up = np.zeros_like(qbit_mapping, dtype=np.bool)
            bit_mask_up[np.where(qbit_mapping & (1 << qbit))] = 1

            amplitude_up = np.linalg.norm(handle_now[bit_mask_up])**2
            amplitude_down = np.linalg.norm(handle_now[~bit_mask_up])**2

            if(amplitude_up > eps):
                handle_next = handle_now.copy()
                handle_next[bit_mask_up] /= np.sqrt(amplitude_up)
                handle_next[~bit_mask_up] = 0
                next_queue.append((prob * amplitude_up, prev_outcome | (1 << qbit), handle_next))
            if(amplitude_down > eps):
                handle_next = handle_now.copy()
                handle_next[~bit_mask_up] /= np.sqrt(amplitude_down)
                handle_next[bit_mask_up] = 0
                next_queue.append((prob * amplitude_down, prev_outcome, handle_next))

    return [[outcome, prob] for prob, outcome, state in next_queue]



import pytest

from pyqcs.gates.builtins import X, H
from pyqcs.gates.circuits import AnonymousCompoundGateCircuit
from pyqcs.state.state import BasicState

def test_or_creates_AnonymousCompoundGateCircuit():
    gc1 = X(0)
    gc2 = X(1)

    circuit = gc1 | gc2

    assert isinstance(circuit, AnonymousCompoundGateCircuit)

def test_or_preserves_order():
    gc1 = X(0)
    gc2 = X(1)

    circuit = gc1 | gc2

    assert circuit.to_executor().to_gate_list() == (gc1.to_executor().to_gate_list() + gc2.to_executor().to_gate_list())
    assert gc1.to_executor().to_gate_list() != gc2.to_executor().to_gate_list()

def test_XH():
    M_SQRT1_2 = 0.70710678118654752440
    state = BasicState.new_zero_state(2)

    state = (X(0) | H(0)) * state

    assert state._qm_state == pytest.approx([M_SQRT1_2, -M_SQRT1_2, 0, 0])



import numpy as np
from .. import X, H, R, C, list_to_circuit


"""
Generate random circuits. Refer to ``random_circuit`` for details.
``random_circuit_XHRC`` uses ``M1=X``, ``M2=H``, ``M3=R``, ``M4=CX``.
"""

def random_basis_values(nqbits, ngates):
    """
    Sample ngates values from {1,...,4 nqbits} x {1,...,nqbits - 1} x [0, 2pi).
    """
    gate_index = np.random.randint(1, high=(nqbits * 4), size=ngates)
    entanglement_index = np.random.randint(1, high=(nqbits - 1), size=ngates)
    phi = np.random.uniform(low=0.0, high=(2 * np.pi), size=ngates)

    return (gate_index, entanglement_index, phi)

def _random_gates(nqbits, ngates, M1, M2, M3, M4):
    for i, k, x in zip(*random_basis_values(nqbits, ngates)):
        if(i <= nqbits):
            yield M1(i - 1)
        elif(i <= 2*nqbits):
            yield M2(i - nqbits - 1)
        elif(i <= 3*nqbits):
            yield M3(i - 2*nqbits - 1, x)

        else:
            if(k <= i - 3*nqbits - 1):
                k -= 1
            yield M4(i - 3*nqbits - 1, k)

def random_circuit(nqbits, ngates, M1, M2, M3, M4):
    """
    Sample ngates gates from (M1, M2, M3, M4) where Mi act on one of the nqbit
    qbits, M1, M2 take no extra arguments (typically M1=X, M2=H), M3 takes a
    float argument (typically M3=R_phi) and M4 is an entanglement gate (typically CX or CZ).

    M1, M2, M3, M4 are equally likely, all qbits are equally likely to be modified.
    Both x for M3 and k for M4 are uniformely distributed.
    """
    return list_to_circuit(list(_random_gates(nqbits, ngates, M1, M2, M3, M4)))

def random_circuit_XHRC(nqbits, ngates):
    return random_circuit(nqbits, ngates, X, H, R, C)




import os
import subprocess
import shutil
from collections import deque
from tempfile import TemporaryDirectory

from numpy import pi

from .flatten import flatten

def not_implemented(*args):
    raise NotImplementedError()


multi_qbit_gates = ["C", "B"]
formatters = {
    "C": lambda c, r: "X"
    , "B": lambda c, r: "Z"
    , "R": lambda c, r: "R_{%.2f\pi}" % (r / pi)
    , "X": lambda c, r: "X"
    , "H": lambda c, r: "H"
    , "Z": lambda c, r: "Z"
    , "GenericGate": not_implemented
    , "M": not_implemented
}

def circuit_to_table(circuit):
    nbits = int(circuit._uses_qbits).bit_length()

    circuits = flatten(circuit)

    rows = [deque() for _ in range(nbits)]
    last = deque()
    for sgc in circuits:
        if(not sgc._descr[0] in multi_qbit_gates):
            can_insert_in_row = True
            for l in last:
                if(l._descr[1] == sgc._descr[1]):
                    can_insert_in_row = False
                    break

            if(can_insert_in_row):
                rows[sgc._descr[1]].append(sgc._descr)
                last.append(sgc)
            else:
                for i,row in enumerate(rows):
                    if(not i in [l._descr[1] for l in last]):
                        row.append(None)
                rows[sgc._descr[1]].append(sgc._descr)
                last = deque([sgc])

        else:
            for i,row in enumerate(rows):
                if(not i in [l._descr[1] for l in last]):
                    row.append(None)
            rows[sgc._descr[1]].append(sgc._descr)
            for i,row in enumerate(rows):
                if(i != sgc._descr[1]):
                    row.append(None)
            last = deque()

    if(last):
        for i,row in enumerate(rows):
            if(not i in [l._descr[1] for l in last]):
                row.append(None)

    return [list(r) for r in rows]


def circuit_to_diagram(circuit):
    table = circuit_to_table(circuit)

    for i,row in enumerate(table):
        for j,format_descriptor in enumerate(row):
            if(format_descriptor is None and table[i][j] is None):
                table[i][j] = r"& \qw"
            if(isinstance(format_descriptor, tuple)):
                if(format_descriptor[0] in multi_qbit_gates):
                    table[format_descriptor[2]][j] = r"& \ctrl{" + str(i - format_descriptor[2]) + r"}"
                table[i][j] = r"& \gate{" + formatters[format_descriptor[0]](*format_descriptor[2:]) + r"}"

    for row in table:
        row.append(r"&\qw")
        row.append(r"\\")

    inner_tex = "\n".join((" ".join(row) for row in table))
    tex = (r"\Qcircuit @C=1em @R=.7em {" + "\n"
            + inner_tex + "\n"
            + r"}")

    return tex

class CircuitPNGFormatter(object):
    def __init__(self, circuit
                    , pdflatex="xelatex"
                    , convert="convert"
                    , pdflatex_args=["main.tex"]
                    , convert_args=["-profile", "\"icc\""
                                , "-density", "300"
                                , "main.pdf"
                                , "-quality", "90"
                                , "main.png"]
                    ):
        self.tex = circuit_to_diagram(circuit)
        self.pdflatex = pdflatex
        self.convert = convert
        self.convert_args = convert_args
        self.pdflatex_args = pdflatex_args

        if(shutil.which(self.pdflatex) is None):
            raise OSError(f"pdflatex ({self.pdflatex}) not found, set it using ``pdflatex=<program>``")
        if(shutil.which(self.convert) is None):
            raise OSError(f"imagemagick ({self.convert}) not found, set it using ``convert=<program>``")

    def _repr_png_(self):
        with TemporaryDirectory() as tmpdirname:
            with open(tmpdirname + "/main.tex", "w") as fout:
                fout.write(self.get_tex_file_content())
            subprocess.run([self.pdflatex] + self.pdflatex_args, cwd=tmpdirname)
            if(not os.path.isfile(tmpdirname + "/main.pdf")):
                raise OSError(f"pdflatex ({self.pdflatex}) did not produce a pdf file")
            subprocess.run([self.convert] + self.convert_args
                            , cwd=tmpdirname)
            if(not os.path.isfile(tmpdirname + "/main.png")):
                raise OSError(f"imagemagick ({self.convert}) did not produce a png file")
            with open(tmpdirname + "/main.png", "rb") as fin:
                return fin.read()

    def get_tex_file_content(self):
        header = r'''
        \documentclass[preview]{standalone}
        \usepackage[utf8]{inputenc}
        \usepackage[T1]{fontenc}

        \usepackage{qcircuit}

        \title{Drawing Circuits with qcircuit}

        \begin{document}
        '''
        bottom = r'''

        \end{document}

        '''

        return header + self.tex + bottom






from ..graph.state import GraphState
from .. import CZ

def graph_lists_to_graph_state(lists, **kwargs):
    vops, entanglements = lists

    state = GraphState.new_plus_state(len(vops), **kwargs)

    handled_enganglements = set()

    for i, neighbors in enumerate(entanglements):
        for j in neighbors:
            edge = tuple(sorted((i, j)))
            if(edge not in handled_enganglements):
                state = CZ(*edge) * state
                handled_enganglements |= {edge}

    for i, vop in enumerate(vops):
        state._g_state.apply_C_L(i, vop)

    return state




class GateListExecutor(object):
    def __init__(self, gate_list):
        self._can_flatten = True
        self._gate_and_executor_list = gate_list
        self.flatten()

    def __call__(self, state):
        for gate_or_executor in self._gate_and_executor_list:
            if(isinstance(gate_or_executor, GateListExecutor)):
                state = gate_or_executor(state)
            else:
                state = state.apply_gate(gate_or_executor)
        return state

    def flatten_generator(self):
        if(not self._can_flatten):
            self.flatten()
            yield self
            return
        for gate_or_executor in self._gate_and_executor_list:
            if(not isinstance(gate_or_executor, GateListExecutor)):
                yield gate_or_executor
            else:
                yield from gate_or_executor.flatten_generator()

    def __inline_flatten_generator(self):
        for gate_or_executor in self._gate_and_executor_list:
            if(not isinstance(gate_or_executor, GateListExecutor)):
                yield gate_or_executor
            else:
                yield from gate_or_executor.flatten_generator()

    def flatten(self):
        self._gate_and_executor_list = [i for i in self.__inline_flatten_generator()]

    def to_gate_list(self):
        return list(self.to_gate_list_generator())

    def to_gate_list_generator(self):
        for gate_or_executor in self._gate_and_executor_list:
            if(not isinstance(gate_or_executor, GateListExecutor)):
                yield gate_or_executor
            else:
                yield from gate_or_executor.to_gate_list_generator()


class RepeatingGateListExecutorSpawner(object):
    def __init__(self, times):
        self._times = times
    def __call__(self, gate_or_executor):
        return RepeatingGateListExecutor(self._times, gate_or_executor)

class RepeatingGateListExecutor(GateListExecutor):
    def __init__(self, times, gate_or_executor):
        self._gate_and_executor_list = gate_or_executor
        self._times = times
        self._can_flatten = False

    def __call__(self, state):
        for _ in range(self._times):
            for gate_or_executor in self._gate_and_executor_list:
                if(isinstance(gate_or_executor, GateListExecutor)):
                    state = gate_or_executor(state)
                else:
                    state = state.apply_gate(gate)
        return state
    def to_gate_list_generator(self):
        for _ in range(self._times):
            for gate_or_executor in self._gate_and_executor_list:
                if(not isinstance(gate_or_executor, GateListExecutor)):
                    yield gate_or_executor
                else:
                    yield from gate_or_executor.to_gate_list_generator()



from .gates.circuits import AnonymousCompoundGateCircuit, NamedCompoundGateCircuit
from .util.to_diagram import CircuitPNGFormatter

def list_to_circuit(list_of_circuits, name=None):
    if(not name):
        return AnonymousCompoundGateCircuit(list_of_circuits)
    return NamedCompoundGateCircuit(list_of_circuits, name, (name,))

def circuitpng(circuit, **kwargs):
    return CircuitPNGFormatter(circuit, **kwargs)



from .abc import AbstractGraphState
from .backend.raw_state import RawGraphState
from .util import graph_lists_to_naive_state
from.gate import GraphGate

class GraphState(AbstractGraphState):
    __slots__ = ["_g_state", "_nbits", "_measured", "_force_new_state"]
    def __init__(self, g_state, nbits, force_new_state=False, measured=None):
        AbstractGraphState.__init__(self)
        self._is_graph = True
        self._g_state = g_state
        self._nbits = nbits
        self._force_new_state = force_new_state
        self._measured = measured
        if(measured is None):
            self._measured = dict()


    @classmethod
    def new_plus_state(cls, nbits, **kwargs):
        if(nbits <= 0):
            raise ValueError("nbits must be greater than 0")

        g_state = RawGraphState(nbits)

        return cls(g_state, nbits, **kwargs)


    def deepcopy(self, **kwargs):
        key_word_arguments = {"force_new_state": self._force_new_state
                            , "measured": self._measured}
        key_word_arguments.update(kwargs)
        return GraphState(self._g_state.deepcopy(), self._nbits, **key_word_arguments)

    def to_naive_state(self):
        return graph_lists_to_naive_state(self._g_state.to_lists())

    def __str__(self):
        return str(self.to_naive_state())

    def check_qbits(self, gate_circuit):
        if(gate_circuit._uses_qbits < (1 << self._nbits)):
            return True
        return False

    @classmethod
    def new_zero_state(cls, nbits, **kwargs):
        if(nbits <= 0):
            raise ValueError("nbits must be greater than 0")

        g_state = RawGraphState(nbits)
        for i in range(nbits):
            g_state.apply_C_L(i, 0)

        return cls(g_state, nbits, **kwargs)

    def apply_gate(self, gate, force_new_state=False):
        if(not isinstance(gate, GraphGate)):
            raise TypeError("gate must be a GraphGate but got {}".format(str(type(gate))))

        state = self
        if(force_new_state or self._force_new_state):
            state = self.deepcopy()

        for op in gate._operation_list:
            has_measured, result = op.apply_to_raw_state(state._g_state)
            if(has_measured):
                state._measured[result[0]] = result[1]

        return state

    def is_normalized(self):
        return True

    def __matmul__(self, other):
        if(not isinstance(other, GraphState)):
            raise TypeError()
        if(not self._nbits == other._nbits):
            raise ValueError("cannot compute overlap of states with different qbit count")
        cmp_state = other._g_state.deepcopy()
        return cmp_state.mul_to(self._g_state)

    def project_to(self, qbit, observable):
        """
        Apply the projection operator of ``observable`` to ``qbit``.
        ``observable`` is a string in ``["Z", "Y", "X", "-Z", "-Y", "-X"]``.
        """
        observables = ["Z", "Y", "X", "-Z", "-Y", "-X"]
        if(observable.upper() not in observables):
            raise ValueError(f"unknown observable, must be one of {observables}")

        observable = observables.index(observable.upper())

        return self._g_state.project_to(qbit, observable)



"""
PyQCS
*****

This package contains a quantum computing simulator.

Notes on Using this Package
###########################

The following gates are imported:

    CX = C(act, control): CNOT
    X(act): NOT, Pauli X gate
    H(act): Hadamard
    R(act, phi): Rotation
    M(act): Measurement
    Z(act): Pauli Z gate
    S(act): S (pi over 4) gate
    CZ(act, control): Controlled Pauli Z gate


The following functions are provided to make measurements easier::

    measure(state, bit_mask) -> (new_state, classical_result)
    sample(state, bit_mask, nsamples, keep_states=False) -> {classical_result: count} or {(new_state, classical_result): count}

Note that ``measure(state, 1)`` is not equivalent to ``M(1) * state``, as
``measure`` will NOT keep previous measurements.  The same applies for
``sample``. Also refer to the section `Imported Utilities`_.

If you need the graphical simulation backend import
``pyqcs.graph.state.GraphState``.


For a quantum simulator state ``state`` one can apply a circuit using the
product::

    new_state = circuit * state

Note that the old state might or might not be modified: If ``state`` is a dense
vector state it will not be affected. If ``state`` is a graphical state and
``force_new_state`` is not specified the old state will be affected.

For a given state class create new states by calling the classmethod
``new_zero_state`` or ``new_plus_state`` (availability depends on the used
state class).

Imported Utilities
##################

``measure``
    Measures once. Refer to function docstring.
``sample``
    Samples from a state. Refer to function docstring.
``tree_amplitudes``
    Computes amplitudes from a state. Refer to function docstring.
``list_to_circuit``
    Converts a list of circuits to a circuit.
``circuitpng``
    Renders a PNG representation of a given circuit in a jupyter notebook.
    Quality of the output varies.

Other Modules
#############

``pyqcs.util.random_circuits``
    Contains a way to generate random circuits. Refer to the module docstring
    for details.
``pyqcs.util.random_graphs``
    The function ``random_graph_state`` creates a graphical state with random
    vertex operators and randomized edges. Refer to the code for details.
``pyqcs.util.to_diagram``
    Contains ``circuit_to_diagram`` which converts a circuit to LaTeX code for
    qcircuit. Quality of the produced code varies.
``pyqcs.util.bytecode``
    Contains ``circuit_to_byte_code`` which converts a Clifford circuit to
    bytecode that can be executed to the pure C graphical simulator. Returns
    a bytecode object. You need to create and add a ``ByteCodeHeader`` to
    execute the bytecode.
``pyqcs.util.flatten``
    Refer to the docstring of function ``flatten``.
``pyqcs.util.from_lists``
    Refer to the code. Mostly for internal use.

Examples
########

See https://github.com/daknuett/PyQCS/tree/master/examples.

"""

import numpy as np

from .gates.builtins import C, H, X, R, M, Z, S, CX, CZ, GenericGate, clear_builtin_gate_cache
from .state.state import BasicState as State
from .gates.circuits import AnonymousCompoundGateCircuit, NamedCompoundGateCircuit


from .measurement import measure, sample, tree_amplitudes
from .utils import list_to_circuit, circuitpng




import pytest
from pyqcs import  M, CZ, CZ, H, R, S, X, State
from pyqcs.graph.state import GraphState
from pyqcs.gates.exceptions import UnitarityError


def test_c_dg_1():
    circuit = ((H(0) | S(1) | H(3) | S(3) | CZ(0, 2) | CZ(1, 3) | H(1) | CZ(0, 2))
                | (X(0) | CZ(0, 1) | H(1) | S(1) | CZ(1, 2)))
    circuit_dagger = circuit.get_dagger()
    state = State.new_zero_state(4)

    s_bar = circuit * state
    s_barbar = circuit_dagger * s_bar

    assert s_barbar == state

def test_c_dg_1g():
    circuit = ((H(0) | S(1) | H(3) | S(3) | CZ(0, 2) | CZ(1, 3) | H(1) | CZ(0, 2))
                | (X(0) | CZ(0, 1) | H(1) | S(1) | CZ(1, 2)))
    circuit_dagger = circuit.get_dagger()
    state = GraphState.new_plus_state(4)

    print(circuit._uses_qbits, circuit._has_graph)
    print(circuit_dagger._uses_qbits, circuit_dagger._has_graph)

    s_bar = circuit * state
    s_barbar = circuit_dagger * s_bar

    assert s_barbar.to_naive_state() == state.to_naive_state()

def test_S_dg():
    circuit = S(0)

    circuit_dagger = circuit.get_dagger()

    assert (circuit | circuit_dagger) * State.new_zero_state(1) == State.new_zero_state(1)

def test_M_dg():
    circuit = H(0) | M(0)

    with pytest.raises(UnitarityError):
        circuit_dagger = circuit.get_dagger()

    assert 1




from pyqcs.gates.circuits import SingleGateCircuit, AnonymousCompoundGateCircuit, NamedCompoundGateCircuit
import pytest

def test_construct_circuit():
    H1 = SingleGateCircuit(1, [], "H", ("H", 1), "H1", None, None)
    X1 = SingleGateCircuit(1, [], "X", ("H", 1), "X1", None, None)
    H2 = SingleGateCircuit(2, [], "H", ("H", 2), "H2", None, None)
    X2 = SingleGateCircuit(2, [], "X", ("H", 2), "X2", None, None)

    assert H1.to_executor().to_gate_list() == ["H1"]
    assert X1.to_executor().to_gate_list() == ["X1"]
    assert H2.to_executor().to_gate_list() == ["H2"]
    assert X2.to_executor().to_gate_list() == ["X2"]

    assert (H1 | H2 | X1 | X2).to_executor().to_gate_list() == ["H1", "H2", "X1", "X2"]

def test_name_constructed_circuit():
    H1 = SingleGateCircuit(1, [], "H", ("H", 1), "H1", None, None)
    X1 = SingleGateCircuit(1, [], "X", ("X", 1), "X1", None, None)
    H2 = SingleGateCircuit(2, [], "H", ("H", 2), "H2", None, None)
    X2 = SingleGateCircuit(2, [], "X", ("X", 2), "X2", None, None)

    c = NamedCompoundGateCircuit.from_anonymous(H1 | H2 | X1 | X2, "HX")

    assert c.to_executor().to_gate_list() == ["H1", "H2", "X1", "X2"]
    assert c._name == "HX"

def test_nested_construct():
    H1 = SingleGateCircuit(1, [], "H", ("H", 1), "H1", None, None)
    X1 = SingleGateCircuit(1, [], "X", ("X", 1), "X1", None, None)
    H2 = SingleGateCircuit(2, [], "H", ("H", 2), "H2", None, None)
    X2 = SingleGateCircuit(2, [], "X", ("X", 2), "X2", None, None)

    assert ((H1 | H2 ) | (X1 | X2)).to_executor().to_gate_list() == ["H1", "H2", "X1", "X2"]



import numpy

from .abc import AbstractGate
from .implementations.basic_gates import BasicGate
from .implementations.generic_gate import GenericGate as _GenericGate


class BaseGate(AbstractGate):
    def __init__(self, impl):
        self._impl = impl

    def __call__(self, qm_state, cl_state):
        return self._impl(qm_state, cl_state)
    def is_inplace(self):
        return False

    def get_dagger(self):
        return BaseGate(self._impl.get_dagger())

class BuiltinGate(BaseGate):
    def __init__(self, type_, act, control, r):
        BaseGate.__init__(self, BasicGate(type_, act, control, r, numpy.random.uniform))

class GenericGate(BaseGate):
    """
    This is the Wrapper for the GenericGate C object.
    """
    def __init__(self, act, arr):
        if(not isinstance(arr, numpy.ndarray)):
            raise TypeError("matrices must be numpy ndarrays")
        if(len(arr.shape) != 2):
            raise ValueError("matrices must be of dimension NxN")
        if(arr.shape[0] != arr.shape[1]):
            raise ValueError("matrices must be of dimension NxN")

        test = arr.dot(arr.conj().T)
        if(not numpy.isclose(test, numpy.identity(arr.shape[0])).all()):
            raise ValueError("matrices must be unitary")

        #if(not numpy.isclose(numpy.linalg.det(arr), 1).all()):
        #    print(numpy.linalg.det(arr))
        #    print(numpy.isclose(numpy.linalg.det(arr), 1))
        #    raise ValueError("matrices must be in SU(N)")

        if(arr.shape[0] != 2):
            raise ValueError("matrices must be in SU(2)")

        arr = arr.astype(numpy.cdouble)

        BaseGate.__init__(self, _GenericGate(act, arr[0,0], arr[1,1], arr[0,1], arr[1,0]))
        self._array = arr

    def get_dagger(self):
        return GenericGate(self._act, self._array.transpose().conjugate())



from abc import ABCMeta, abstractmethod
import logging

logger = logging.getLogger(__name__)

try:
    import ray
except ImportError:
    raise ImportError("pyqcs experiments require ray for parallelization; install it using pip3 install ray")


class Instruction(metaclass=ABCMeta):
    @abstractmethod
    def __call__(self, params):
        pass

class FunctionInstruction(Instruction):
    def __init__(self, name, fct):
        self.name = name
        self.fct = fct
    def __call__(self, params):
        logger.info(f"executing instruction {self.name}")
        return self.fct(params)


@ray.remote
class Workflow(object):
    def __init__(self, name, instructions):
        self.name = name
        self._instructions = instructions

    def execute(self, params):
        logger.info(f"executing workflow {self.name}")
        state = params
        for i,instr in enumerate(self._instructions):
            logger.info(f"workflow {self.name}: instruction {i}")
            state = instr(state)
        return state

class WorkflowSpawner(object):
    def __init__(self, name, instructions):
        self.name = name
        self.instructions = instructions
    def spawn(self):
        return Workflow.remote(self.name, self.instructions)





from .abc import AbstractGateCircuit, AbstractNamedGateCircuit, AbstractCompoundGateCircuit
from .executor import GateListExecutor

class SingleGateCircuit(AbstractNamedGateCircuit):
    def __init__(self
            , qbits
            , identities
            , name
            , descr
            , gate_naive
            , gate_graph
            , make_dagger):
        AbstractNamedGateCircuit.__init__(self, qbits, identities, name, descr)
        if(gate_naive is not None):
            self._has_naive = True
        self._gate = gate_naive
        if(gate_graph is not None):
            self._has_graph = True
        self._gate_g = gate_graph

        self._make_dagger = make_dagger


    def get_child_executors(self, graph=False):
        if(not graph):
            return [GateListExecutor([self._gate])]
        else:
            return [GateListExecutor([self._gate_g])]

    def gate_list_generator(self):
        yield self._gate
    def __or__(self, other):
        if(not isinstance(other, AbstractGateCircuit)):
            raise TypeError()
        return AnonymousCompoundGateCircuit([self, other])

    def __ror__(self, other):
        if(not isinstance(other, AbstractGateCircuit)):
            raise TypeError()
        return AnonymousCompoundGateCircuit([other, self])


    def new_from_circuit_with_executor(self, executor):
        new_circuit = AnonymousCompoundGateCircuit([self], qbits=self._uses_qbits
                                                    , has_graph=self._has_graph
                                                    , has_naive=self._has_naive)
        new_circuit._executor = executor
        return new_circuit

    def get_dagger(self):
        return self._make_dagger()


class AnonymousCompoundGateCircuit(AbstractCompoundGateCircuit):
    def __init__(self, subcircuit_list, has_graph=None, has_naive=None, qbits=None):
        if(qbits is None):
            qbits = 0
            for subcircuit in subcircuit_list:
                qbits |= subcircuit._uses_qbits

        AbstractCompoundGateCircuit.__init__(self, qbits, [], subcircuit_list)

        if(has_graph is None or has_naive is None):
            has_graph = True
            has_naive = True
            for subc in subcircuit_list:
                has_graph = has_graph and subc._has_graph
                has_naive = has_naive and subc._has_naive

        self._has_naive = has_naive
        self._has_graph = has_graph


    def __or__(self, other):
        if(not isinstance(other, AbstractGateCircuit)):
            raise TypeError()
        return AnonymousCompoundGateCircuit([self, other])

    def __ror__(self, other):
        if(not isinstance(other, AbstractGateCircuit)):
            raise TypeError()
        return AnonymousCompoundGateCircuit([other, self])

    def new_from_circuit_with_executor(self, executor):
        new_circuit = AnonymousCompoundGateCircuit(self._subcircuits, qbits=self._uses_qbits)
        new_circuit._executor = executor
        new_circuit._has_graph = self._has_graph
        new_circuit._has_naive = self._has_naive
        return new_circuit

    def get_child_executors(self, graph=False):
        return [c.to_executor(graph=graph) for c in self._subcircuits]

    def get_dagger(self):
        return AnonymousCompoundGateCircuit([s.get_dagger() for s in reversed(self._subcircuits)]
                                            , has_graph=self._has_graph
                                            , has_naive=self._has_naive
                                            , qbits=self._uses_qbits)

class NamedCompoundGateCircuit(AnonymousCompoundGateCircuit, AbstractNamedGateCircuit):
    def __init__(self, subcircuit_list, name, descr, identities=[]):
        qbits = 0
        for subcircuit in subcircuit_list:
            qbits |= subcircuit._uses_qbits
        has_graph = True
        has_naive = True
        for subc in subcircuit_list:
            has_graph = has_graph and subc._has_graph
            has_naive = has_naive and subc._has_naive

        AnonymousCompoundGateCircuit.__init__(self, subcircuit_list, has_graph=has_graph, has_naive=has_naive, qbits=qbits)
        AbstractNamedGateCircuit.__init__(self, qbits, identities, name, descr)

    @classmethod
    def from_anonymous(cls, anonymous, name, identities=[]):
        if(not isinstance(anonymous, AnonymousCompoundGateCircuit)):
            raise TypeError("anonymous must be of type AnonymousCompoundGateCircuit")
        subcircuit_list = anonymous._subcircuits
        return cls(subcircuit_list, name, identities)


    def new_from_circuit_with_executor(self, executor):
        # Don't give that new circuit a name. It is a new
        # circuit.
        new_circuit = AnonymousCompoundGateCircuit([self], self._uses_qbits)
        new_circuit._executor = executor
        return new_circuit




from abc import abstractmethod, ABCMeta
from .backend.raw_state import RawGraphState
from numpy.random import uniform

class AbstractGraphOperation(metaclass=ABCMeta):
    def __init__(self, act):
        self._act = act

    @abstractmethod
    def apply_to_raw_state(self, state):
        pass

class CZOperation(AbstractGraphOperation):
    def __init__(self, act, control):
        AbstractGraphOperation.__init__(self, act)
        self._control = control

    def apply_to_raw_state(self, state):
        if(not isinstance(state, RawGraphState)):
            raise TypeError("state must be of type RawGraphState, but got {}".format(str(type(state))))

        state.apply_CZ(self._act, self._control)
        return False, None

class CLOperation(AbstractGraphOperation):
    def __init__(self, act, clifford_index):
        AbstractGraphOperation.__init__(self, act)

        if(not isinstance(clifford_index, int)):
            raise TypeError("clifford_index must be integer")
        if(0 > clifford_index > 23):
            raise ValueError("clifford_index must be between 0 and 23")
        self._clifford_index = clifford_index

    def apply_to_raw_state(self, state):
        if(not isinstance(state, RawGraphState)):
            raise TypeError("state must be of type RawGraphState, but got {}".format(str(type(state))))

        state.apply_C_L(self._act, self._clifford_index)
        return False, None

class MeasurementOperation(AbstractGraphOperation):
    def __init__(self, act):
        AbstractGraphOperation.__init__(self, act)

    def apply_to_raw_state(self, state):
        if(not isinstance(state, RawGraphState)):
            raise TypeError("state must be of type RawGraphState, but got {}".format(str(type(state))))

        rnd = uniform()
        result = state.measure(self._act, rnd)

        return True, (self._act, result)


class GraphGate(object):
    def __init__(self, operation_list):
        self._operation_list = operation_list



from collections import deque
import numpy as np

try:
    import ray
except ImportError:
    raise ImportError("pyqcs experiments require ray for parallelization; install it using pip3 install ray")

def approximate_gradient(pool, x, eps):
    """
    Compute an appoximate gradient using

    .. math::

        F'(x) \\approx \\frac{F(x + \\epsilon) - F(x - \\epsilon)}{2\\epsilon}

    ``pool`` is an ``ActorPool`` that evaluates the objective
    function.
    """
    evaluation_points = deque()
    for i,_ in enumerate(x):
        ei = np.zeros_like(x, dtype=np.double)
        ei[i] = eps
        evaluation_points.append(x + ei)
        evaluation_points.append(x - ei)

    data_points = list(pool.map(lambda a,v: a.execute.remote(v), evaluation_points))

    gradient = np.array([(data_points[2*i] - data_points[2*i + 1])/(2*eps) for i,_ in enumerate(x)])
    return gradient

def approx_gradient_decent(wf_spawner, x0, eps, gamma, nmax=40_000, rho=1e-6):
    """
    Perform a gradient decent using an appoximate gradient (computed
    by ``approximate_gradient``; ``rho`` is passed to approximate_gradient as
    parameter ``eps``).

    ``wf_spawner`` is a ``WorkflowSpawner`` used to compute the objective function.

    ``x0`` is the starting point, ``eps`` is the minimal norm of the gradient to
    continue, ``gamma`` is the step width parameter, and ``nmax`` specifies how
    many steps are made at most.

    Returns ``success: bool, x_final``.

    Uses an ``ActorPool`` with ``len(x0)`` actors.

    This function expects the objective function to be costly.
    Therefore all function evaluations are parallelized.
    """
    pool = ray.util.ActorPool([wf_spawner.spawn() for _ in x0])
    x = x0

    for n in range(nmax):
        gradient = approximate_gradient(pool, x, rho)
        if(np.linalg.norm(gradient) < eps):
            return True, x

        x = x - gamma*gradient

    if(np.linalg.norm(gradient) < eps):
        return True, x
    return False, x




import numpy as np
from pyqcs import H, CX, State, tree_amplitudes

def test_tree_amplitudes_plus_state():
    state = (H(0) | H(1) | H(2) | H(3)) * State.new_zero_state(4)

    outcome = tree_amplitudes(state)

    assert np.allclose(list(sorted(outcome, key=lambda x: x[0])), [[i, 0.5**4] for i in range(2**4)])

def test_tree_amplitudes_double_bell_state():
    state = (H(0) | CX(1, 0) | H(2) | CX(3, 2)) * State.new_zero_state(4)

    outcome = tree_amplitudes(state)

    assert np.allclose(outcome, [[0b0000, 0.2500000000000001],
                         [0b1100, 0.2500000000000001],
                         [0b0011, 0.2500000000000001],
                         [0b1111, 0.2500000000000001]])



from pyqcs import H, S, clear_builtin_gate_cache
from pyqcs.gates.builtins import _H, _S


def test_cache_is_active():
    gates_S = [S(i) for i in range(10)]
    gates_H = [H(i) for i in range(20)]

    for s in gates_S:
        assert s in _S._registry.values()

    for h in gates_H:
        assert h in _H._registry.values()

def test_clear_cache():
    gates_S = [S(i) for i in range(10)]
    gates_H = [H(i) for i in range(20)]

    clear_builtin_gate_cache()

    assert _S._registry == {}
    assert _H._registry == {}



import numpy as np
from .gate import BuiltinGate, GenericGate as _GenericGate
from .circuits import SingleGateCircuit
from .exceptions import UnitarityError
from ..build.abc import AbstractSingleGateCircuitBuilder
from ..graph.gate import GraphGate, CLOperation, CZOperation, MeasurementOperation

class BuiltinGateBuilder(AbstractSingleGateCircuitBuilder):
    __slots__ = ["_type"]
    def __init__(self, type_, get_graph_gate, has_control=False, can_dagger=True):
        AbstractSingleGateCircuitBuilder.__init__(self)
        self._type = type_
        self.get_graph_gate = get_graph_gate
        self._has_control = has_control
        self._can_dagger = can_dagger

    def __call__(self, act, *args, dagger=False):
        # This fixes a rare bug when generating circuits
        # with NumPy. When more qbits are used than
        # the NumPy fixed width integer can hold one can
        # experience weird behaviour
        act = int(act)
        args = list(args)
        args[0] = int(args[0])

        if(dagger):
            args = (args[0], -args[1])

        if((act, *args) not in self._registry):
            gate = BuiltinGate(self._type, act, *args)
            gate_graph = self.get_graph_gate(act, *args, dagger=dagger)

            uses_qbits = (1 << act)
            if(self._has_control):
                uses_qbits |= (1 << args[0])

            def mk_dg():
                if(self._can_dagger):
                    return self(act, *args, dagger=True)
                else:
                    raise UnitarityError("gate is not unitary and cannot be daggered")

            circuit = SingleGateCircuit(uses_qbits
                        , []
                        , self._type + "(" + ",".join((str(a) for a in args)) + ")"
                        , (self._type, act, *args)
                        , gate
                        , gate_graph
                        , mk_dg)
            self._registry[(act, *args)] = circuit
        return self._registry[(act, *args)]

def _get_graph_H_gate(act, i1, i2, dagger=False):
    return GraphGate([CLOperation(act, 0)])
_H = BuiltinGateBuilder('H', _get_graph_H_gate)
def H(act):
    if(act < 0):
        raise ValueError("act qbit must be >= 0")
    return _H(act, 0, 0)

def _get_graph_X_gate(act, i1, i2, dagger=False):
    return GraphGate([CLOperation(act, 14)])
_X = BuiltinGateBuilder('X', _get_graph_X_gate)
def X(act):
    if(act < 0):
        raise ValueError("act qbit must be >= 0")
    return _X(act, 0, 0)


def _get_graph_M_gate(act, i1, i2, dagger=False):
    return GraphGate([MeasurementOperation(act)])
_M = BuiltinGateBuilder('M', _get_graph_M_gate, can_dagger=False)
def M(act):
    if(act < 0):
        raise ValueError("act qbit must be >= 0")
    return _M(act, 0, 0)

def _get_graph_C_gate(act, control, i2, dagger=False):
    return GraphGate([CLOperation(act, 0)
                    , CZOperation(act, control)
                    , CLOperation(act, 0)])
_C = BuiltinGateBuilder('C', _get_graph_C_gate, has_control=True)
def C(act, control):
    if(act < 0):
        raise ValueError("act qbit must be >= 0")
    if(act == control):
        raise ValueError("act qbit and control qbit must be different")
    return _C(act, control, 0)

CX = C
CNOT = C

def _get_graph_R_gate(i0, i1, i2, dagger=False):
    return None
_R = BuiltinGateBuilder('R', _get_graph_R_gate)
def R(act, r):
    if(act < 0):
        raise ValueError("act qbit must be >= 0")
    return _R(act, 0, r)

def _get_graph_Z_gate(act, i1, i2, dagger=False):
    return GraphGate([CLOperation(act, 5)])
_Z = BuiltinGateBuilder('Z', _get_graph_Z_gate)
def Z(act):
    if(act < 0):
        raise ValueError("act qbit must be >= 0")
    return _Z(act, 0, 0)

def _get_graph_B_gate(act, control, i2, dagger=False):
    return GraphGate([CZOperation(act, control)])
_B = BuiltinGateBuilder('B', _get_graph_B_gate, has_control=True)
def CZ(act, control):
    if(act < 0):
        raise ValueError("act qbit must be >= 0")
    if(act == control):
        raise ValueError("act qbit and control qbit must be different")
    return _B(act, control, 0)

def _get_graph_S_gate(act, i1, i2, dagger=False):
    if(not dagger):
        return GraphGate([CLOperation(act, 1)])
    else:
        return GraphGate([CLOperation(act, 8)])

_S = BuiltinGateBuilder('R', _get_graph_S_gate)
def S(act):
    if(act < 0):
        raise ValueError("act qbit must be >= 0")
    return _S(act, 0, np.pi/2)


def GenericGate(act, array):
    """
    This is the circuit builder that is used to generate 1-qbit
    generic gates.
    """
    if(act < 0):
        raise ValueError("act qbit must be >= 0")
    gate = _GenericGate(act, array)

    circuit = SingleGateCircuit(1 << act
            , []
            , str(array)
            , ("GenericGate", act, array)
            , gate
            , None
            , None)
    def make_dagger():
        return SingleGateCircuit(1 << act
            , []
            , str(array.transpose().conjugate())
            , ("GenericGate", act, array.transpose().conjugate())
            , gate
            , None
            , lambda: circuit)

    circuit._make_dagger = make_dagger
    return circuit

def clear_builtin_gate_cache():
    _H.clear_gate_registry()
    _X.clear_gate_registry()
    _C.clear_gate_registry()
    _M.clear_gate_registry()
    _S.clear_gate_registry()
    _B.clear_gate_registry()
    _Z.clear_gate_registry()
    _R.clear_gate_registry()




from abc import ( ABCMeta
                , abstractmethod)


from .executor import GateListExecutor, RepeatingGateListExecutorSpawner
from ..state.abc import AbstractState
from ..graph.abc import AbstractGraphState

class AbstractGateCircuit(metaclass=ABCMeta):
    def __init__(self, qbits, identities):
        self._has_graph = False
        self._has_naive = False
        self._uses_qbits = qbits
        self._identities = identities
        self._executor = GateListExecutor

    def __mul__(self, other):
        if(isinstance(other, AbstractGraphState)):
            if(not self._has_graph):
                raise TypeError("Cannot apply circuit to graph state. Check your gates.")
            if(not other.check_qbits(self)):
                raise ValueError(
                        "Gate circuit requires qbits {}, but given state has less.".format(
                            bin(self._uses_qbits)
                            ))
            return self._executor(self.get_child_executors(graph=True))(other)
        if(isinstance(other, AbstractState)):
            if(not other.check_qbits(self)):
                raise ValueError(
                        "Gate circuit requires qbits {}, but given state has less.".format(
                            bin(self._uses_qbits)
                            ))
            return self._executor(self.get_child_executors())(other)
        if(isinstance(other, int)):
            return self.new_from_circuit_with_executor(RepeatingGateListExecutorSpawner(other))
        raise TypeError()

    def __rmul__(self, other):
        if(isinstance(other, int)):
            return self.new_from_circuit_with_executor(RepeatingGateListExecutorSpawner(other))
        raise TypeError()


    def add_identity(self, identity):
        self._identities.append(identity)

    def to_executor(self, graph=False):
        return self._executor(self.get_child_executors(graph=graph))

    @abstractmethod
    def __or__(self, other):
        pass

    @abstractmethod
    def __ror__(self, other):
        pass

    @abstractmethod
    def get_child_executors(self, graph=False):
        pass

    @abstractmethod
    def new_from_circuit_with_executor(self, executor):
        pass

    @abstractmethod
    def get_dagger(self):
        pass

class AbstractNamedGateCircuit(AbstractGateCircuit):
    def __init__(self, qbits, identities, name, descr):
        AbstractGateCircuit.__init__(self, qbits, identities)
        self._name = name
        self._descr = descr

class AbstractCompoundGateCircuit(AbstractGateCircuit):
    __slots__ = ["_subcircuits"]
    def __init__(self, qbits, identities, subcircuits):
        AbstractGateCircuit.__init__(self, qbits, identities)
        self._subcircuits = subcircuits


class AbstractGate(metaclass=ABCMeta):
    @abstractmethod
    def __call__(self, qm_state, cl_state):
        pass
    @abstractmethod
    def get_dagger(self):
        pass




import pytest

from pyqcs import State, C, H, X, R

@pytest.fixture
def zero_state():
    return State.new_zero_state(2)

def test_repeating_gates():
    base = X(0) | H(0)

    multiplied = base * 5

    assert multiplied.to_executor().to_gate_list() == (5 * base.to_executor().to_gate_list())

def test_repeating_gates_inverted():
    base = X(0) | H(0)

    multiplied = 5 * base

    assert multiplied.to_executor().to_gate_list() == (5 * base.to_executor().to_gate_list())



import struct
from collections import deque

from ..gates.builtins import CLOperation, CZOperation, MeasurementOperation


class ByteCodeHeader(object):
    def __init__(self, nqbits, sample, nsamples):

        if(isinstance(sample, int)):
            self._sample = [i for i in range(sample.bit_length()) if sample & (1 << i)]
            self._sample_qbit_count = len(self._sample)
        elif(isinstance(sample, (tuple, list))):
            for i in sample:
                if(not isinstance(i, int)):
                    raise TypeError("sample must be int or list of ints")
            self._sample = sample
            self._sample_qbit_count = len(self._sample)
        else:
            raise TypeError("sample must be int or list of ints")

        self._nsamples = nsamples
        self._nqbits = nqbits

        if(max(self._sample) >= self._nqbits):
            raise ValueError("one qbit to sample is out of qbit range")
        if(min(self._sample) < 0):
            raise ValueError("one qbit to sample is out of qbit range")

    def to_bytes(self):
        magic_word = b"GQCS"

        b = struct.pack("Q", self._nqbits)
        s = struct.pack("H", self._nsamples)
        q = struct.pack("Q", self._sample_qbit_count)

        samples = [struct.pack("Q", i) for i in self._sample]

        return magic_word + b + b"b" + s + b"s" + q + b"q" + b"".join(samples) + b"\xff"*8

class ByteCodeInstruction(object):
    cmds = {"M": b"M", "CZ": b"Z", "C_L": b"L"}

    def __init__(self, command, act, argument):
        self._command = self.cmds[command]
        self._act = act
        self._argument = argument

    def to_bytes(self):
        return self._command + struct.pack("Q", self._act) + struct.pack("Q", self._argument)

def circuit_to_byte_code(header, circuit, add_header=True):
    if(not circuit._has_graph):
        raise TypeError("Circuit must be applicable to graph.")
    if(circuit._uses_qbits >= (1 << header._nqbits)):
        raise ValueError("Circuit requires more qbits.")

    executor = circuit._executor(circuit.get_child_executors(graph=True))
    gates = executor.to_gate_list()

    def C_L_to_instruction(op):
        return ByteCodeInstruction("C_L", op._act, op._clifford_index)
    def CZ_to_instruction(op):
        return ByteCodeInstruction("CZ", op._act, op._control)
    def M_to_instruction(op):
        return ByteCodeInstruction("M", op._act, 0)

    o2i = {CLOperation: C_L_to_instruction
            , CZOperation: CZ_to_instruction
            , MeasurementOperation: M_to_instruction}

    instructions = deque()
    for gate in gates:
        for op in gate._operation_list:
            instructions.append(o2i[type(op)](op))

    if(add_header):
        result = header.to_bytes()
    else:
        result = b""

    result += b"".join((i.to_bytes() for i in instructions))
    return result



from ..gates.circuits import SingleGateCircuit, AbstractCompoundGateCircuit

def _flatten_generator(circuit):
    if(isinstance(circuit, SingleGateCircuit)):
        yield circuit
        return

    if(isinstance(circuit, AbstractCompoundGateCircuit)):
        for subc in circuit._subcircuits:
            yield from _flatten_generator(subc)
        return

    raise ValueError(f"cannot flatten {type(circuit)}")

def flatten(circuit):
    """
    Flattens a (possibly compound) circuit to a list of SingleGateCircuits.
    """
    return list(_flatten_generator(circuit))



"""
This module contains just a lookup table for the local Clifford group.
"""

C_L = ['H'
    , 'S'
    , 'ZZ'
    , 'SH'
    , 'HS'
    , 'Z'
    , 'SHS'
    , 'HZ'
    , 'ZS'
    , 'SXH'
    , 'HSZ'
    , 'HXSH'
    , 'HSH'
    , 'HX'
    , 'X'
    , 'ZSH'
    , 'SX'
    , 'HXS'
    , 'XS'
    , 'SZHS'
    , 'ZXH'
    , 'XZ'
    , 'XSH'
    , 'HSX']



from abc import ( ABCMeta
        , abstractclassmethod
        , abstractmethod)

class AbstractState(metaclass=ABCMeta):
    __slots__ = ["_is_graph", "_is_naive"]
    def __init__(self):
        self._is_graph = False
        self._is_naive = False

    @abstractmethod
    def apply_gate(self, gate, force_new_state=False):
        pass
    @abstractmethod
    def __str__(self):
        pass
    @abstractmethod
    def check_qbits(self, gate_circuit):
        pass
    @abstractmethod
    def deepcopy(self, **kwargs):
        pass

    @abstractclassmethod
    def new_zero_state(cls, nbits, **kwargs):
        pass
    @abstractmethod
    def is_normalized(self):
        pass



from setuptools import setup, Extension, find_packages
import numpy


basic_gates = Extension(
                        "pyqcs.gates.implementations.basic_gates"
                        , sources=["src/pyqcs/gates/implementations/basic_gates.c"]
                        , extra_compile_args=["-fstack-protector"
                                            , "-Wno-unused-variable"
                                            , "-g"
                                            , "-I%s" % numpy.get_include()]
                    )
graph_backend = Extension("pyqcs.graph.backend.raw_state"
                        , sources=["src/pyqcs/graph/backend/raw_state.c"
                                , "src/pyqcs/graph/backend/linked_list.c"
                                , "src/pyqcs/graph/backend/graph_operations.c"
                        ]
                        , extra_compile_args=["-fstack-protector"
                                             , "-Wno-unused-variable"
                                             , "-g"
                                             , "-I%s" % numpy.get_include()]
                    )
generic_gate = Extension(
                        "pyqcs.gates.implementations.generic_gate"
                        , sources=["src/pyqcs/gates/implementations/generic_gate.c"]
                        , extra_compile_args=["-fstack-protector"
                                            , "-Wno-unused-variable"
                                            , "-g"
                                            , "-I%s" % numpy.get_include()])

setup(
        name="pyqcs"
        , version="2.5.0"
        , description="A quantum computing simulator."
        , long_description = open("README.rst").read()
        , ext_modules=[basic_gates
                        , generic_gate
                        , graph_backend]
        , packages=find_packages(where="src")
        , package_dir={"pyqcs": "src/pyqcs"}
        , install_requires=["numpy", "ray"]
        , project_urls={
            "Source Code": "https://github.com/daknuett/pyqcs"
        }
        , url="https://github.com/daknuett/pyqcs"
        , author="Daniel Knüttel"
        , author_email="daniel.knuettel@daknuett.eu"
        , python_requires=">3.5"
        , classifiers=[
            "License :: OSI Approved :: GNU General Public License v3 (GPLv3)"
        ]
        , license="GNU General Public License v3 (GPLv3)"
    )




from .abc import AbstractState
import numpy as np

class BasicState(AbstractState):
    """
    This is the basic dense state vector class. It uses NumPy Arrays to
    store the coefficients. There is no compression. Gates are implemented
    as NumPy UFuncs.

    Use ``BasicState.new_zero_state(nqbits: int)`` to instantiate new states.

    It is possible to compute the overlap of two states using ``state1 @ state2
    -> complex``. Use ``state1 == state2`` to compare states, however, this
    will disregard a global phase.

    """
    __slots__ = ["_nbits", "_ndim", "_qm_state", "_cl_state", "_last_measured"
            , "_length_error", "_check_normalization"]
    def __init__(self
                , qm_state
                , cl_state
                , nbits
                , last_measured
                , lenght_error=1e-8
                , check_normalization=False
                , **kwargs):
        AbstractState.__init__(self)
        self._is_naive = True
        self._qm_state = qm_state
        self._cl_state = cl_state
        self._nbits = nbits
        self._ndim = 2**nbits
        self._last_measured = last_measured
        self._length_error = lenght_error
        self._check_normalization = check_normalization

        if(check_normalization and not self.is_normalized()):
            raise ValueError("State is not normalized")

    @classmethod
    def new_zero_state(cls, nbits, **kwargs):
        if(nbits > 29):
            # (2**29 * 16) / 2**30 = 8 GiB.
            raise Warning("State will require more than 8GiB RAM. "
                        "Using such big states can lead to random crashes.")
        ndim = 2**nbits
        qm_state = np.zeros(ndim, dtype=np.cfloat)
        qm_state[0] = 1
        last_measured = 0
        cl_state = -1 * np.ones(nbits, dtype=np.int8)

        return cls(qm_state, cl_state, nbits, last_measured, **kwargs)

    def check_qbits(self, gate_circuit):
        if(gate_circuit._uses_qbits < (1 << self._nbits)):
            return True
        return False

    def deepcopy(self, **kwargs):
        keyword_arguments = {"check_normalization": self._check_normalization
                            , "lenght_error": self._length_error}
        keyword_arguments.update(kwargs)
        return BasicState(self._qm_state.copy()
                        , self._cl_state.copy()
                        , self._nbits
                        , self._last_measured
                        , **keyword_arguments)

    def apply_gate(self, gate, force_new_state=False):
        qm_state, cl_state, last_measured = gate(self._qm_state, self._cl_state)
        return BasicState(qm_state
                        , cl_state
                        , self._nbits
                        ,  last_measured
                        , lenght_error=self._length_error
                        , check_normalization=self._check_normalization)

    def _easy_format_state_part(self, cf, i):
        return "{}*|{}>".format(str(cf), bin(i))

    def __str__(self):
        eps = 1e-13

        s = " + ".join((self._easy_format_state_part(self._qm_state[i], i)
                        for i in range(self._ndim) if (abs(self._qm_state[i]) > eps)))
        return s

    def __repr__(self):
        return str(self)

    def __hash__(self):
        return hash((tuple(self._qm_state), tuple(self._cl_state)))

    def __eq__(self, other):
        if(isinstance(other, BasicState)):
            if(self._nbits != other._nbits):
                return False
            if(not np.allclose(self._cl_state, other._cl_state)):
                return False

            overlap = self @ other
            if(not np.isclose(np.absolute(overlap), 1)):
                return False
            return True

        raise TypeError()


    def __matmul__(self, other):
        if(not isinstance(other, BasicState)):
            raise TypeError()
        if(not (other.is_normalized() and self.is_normalized())):
            raise ValueError("states must be normalized")
        if(self._nbits != other._nbits):
            raise ValueError("states must have same qbit count")

        return self._qm_state.conjugate().dot(other._qm_state)

    def is_normalized(self):
        return np.isclose(np.sum(np.absolute(self._qm_state)**2), 1, atol=self._length_error)

    def projZ(self, m, l):
        """
        Applies

        ..math::
                P_{m,l} = \\frac{I + (-1)^l Z_m}{2}

        to the state. ``l`` must be ``0`` or ``1``.

        Returns either ``0`` (if the amplitude is below self._length_error)
        or a new state (in any other case).
        """
        if l not in (0,1):
            raise ValueError("l must be 0 or 1")
        if(m >= self._nbits):
            raise ValueError(f"qbit m out of range (0, ..., {self._nbits - 1})")

        indices = np.arange(0, self._ndim, 1, dtype=np.int)

        indices[indices & (1 << m) == 0] = 0
        indices[indices != 0] = 1
        indices = indices.astype(np.bool)

        qm_state = self._qm_state.copy()
        if(not l):
            qm_state[indices] = 0
        else:
            qm_state[~indices] = 0

        amplitude = np.linalg.norm(qm_state)
        if(amplitude < self._length_error):
            return 0

        qm_state /= amplitude

        state = self.deepcopy()
        state._qm_state = qm_state

        return state



import pytest

from pyqcs import State, X, H, C, R, M

def test_missing_qbit1():
    state = State.new_zero_state(2)

    with pytest.raises(ValueError):
        assert X(2) * state

def test_missing_qbitmany():
    state = State.new_zero_state(2)

    with pytest.raises(ValueError):
        assert X(200) * state



from numpy.random import choice

from itertools import product

from .from_lists import graph_lists_to_graph_state

def random_twoclique_graph(nother, n0):
    if(n0 > nother):
        raise ValueError("n0 must be <= nother")
    V_bar = [i + 2 for i in range(nother)]
    P_0 = V_bar[:n0]
    P_1 = V_bar[n0:]
    A = [0]
    B = [1]

    E = set()

    for i in P_0:
        a = choice(A)
        E |= {frozenset((i, a))}
        A.append(i)

    for i in P_1:
        b = choice(B)
        E |= {frozenset((i, b))}
        B.append(i)

    V = set(V_bar) | {0, 1}

    return V, E

def random_nobipartite_graph(nother, n0, naddedges):
    V, E = random_twoclique_graph(nother, n0)
    E_m = {frozenset((i,j)) for i,j in product(V, V) if i != j} - E
    if(naddedges < 0
            or naddedges > len(E_m)):
        raise ValueError(f"naddedges must be > 0 and < {len(E_m)}")

    N = set(choice(list(E_m), size=naddedges))

    return V, E | N


def random_graph_lists(nqbits, naddedges):
    if(nqbits < 4):
        raise ValueError("at least 4 qbits are required for a meaningful random graph")
    n0 = choice(nqbits - 2)
    vops = choice(24, size=nqbits)

    V, E = random_nobipartite_graph(nqbits - 2, n0, naddedges)

    ngbhds = []
    for v in V:
        ngbhds.append(set())
        for i,j in E:
            if(i == v):
                ngbhds[v] |= {j}
            if(j == v):
                ngbhds[v] |= {i}

    return vops, [list(sorted(n)) for n in ngbhds]

def random_graph_state(nqbits, naddedges):
    return graph_lists_to_graph_state(random_graph_lists(nqbits, naddedges))
